#ifndef LIB_ERROR_HANDLER_DEFINITIONS_MQH_INCLUDED
#define LIB_ERROR_HANDLER_DEFINITIONS_MQH_INCLUDED
#property version "4.2";
/**********************************************************************************
 * Copyright (C) 2010-2022 Dominik Egert <info@freie-netze.de>
 *
 * This file is error code resolver definitions include file.
 *
 * MQLplus, including this file may not be copied and/or distributed 
 * without explecit permit by the author.
 * Author Dominik Egert / Freie Netze UG.
 **********************************************************************************
 *
 *  Version: 4.2
 *  State: public
 *
 *  File information
 *  ================
 *
 *  Use: resolve error codes to text
 *
 *
 *
*/
#ifdef DBG_MSG_TRACE_FILE_LOADER
    DBG_MSG_TRACE_FILE_LOADER;

#endif


//*********************************************************************************************************************************************************/
// Global error handler configuration
//

////////////////////////////////////////////////////
//
// Define error group id
//  This ID must be unique within
//  the given project.
//
//  Valid values for Group ID are
//  between 0x00 and 0xEF
//  0xF0 to 0xFF is reserved.
//
//  Error codes categories
//      0x00            = Succes
//      0x01 ... 0xDF   = Errors,
//      0xE0 ... 0xEF   = Warnings
//      0xF0 ... 0xFF   = Handled Errors
//
//  Define the size each error
//  router can handle. (0x0100 = 256 here)
//
#ifndef LIB_ERR_LEVEL_WARNING
    #define LIB_ERR_LEVEL_WARNING       0x00E0

#endif
#ifndef LIB_ERR_LEVEL_HANDLED_ERROR
    #define LIB_ERR_LEVEL_HANDLED_ERROR 0x00F0

#endif
#ifndef LIB_ERR_INTERNAL_GROUP_IDS
    #define LIB_ERR_INTERNAL_GROUP_IDS  0xF000

#endif
#ifndef LIB_ERR_GROUP_SIZE
    #define LIB_ERR_GROUP_SIZE          0x0100

#endif

//
////////////////////////////////////////////////////



////////////////////////////////////////////////////
//
// Define error messages
//  output format
//

// Define error output format
#ifndef LIB_ERR_OUTPUT_FORMAT
    #define LIB_ERR_OUTPUT_FORMAT(_err_code_, _code_name_, _msg_) StringFormat("(%i - %s) %s", _err_code_, _code_name_, _msg_)

#endif

//
////////////////////////////////////////////////////

//
// END Global error handler configuration
//*********************************************************************************************************************************************************/





/////////////////////////////////////////////////////////////////////////////////////////////////////
//
// Configure debug mode
//  Do not use shared library in debugging if not forced
//
//  Optionally force shared liibrary
//      #define LIB_ERROR_FORCE_SHARED_LIBRARY_IMPORT
//
#ifdef LIB_DEBUG
    #ifdef LIB_ERROR_IMPORT_SHARED_LIBRARY
        #ifndef LIB_ERROR_FORCE_SHARED_LIBRARY_IMPORT
            #undef LIB_ERROR_IMPORT_SHARED_LIBRARY

        #endif

    #endif 
    #ifdef LIB_ERROR_FORCE_SHARED_LIBRARY_IMPORT
        #ifndef LIB_ERROR_IMPORT_SHARED_LIBRARY
            #define LIB_ERROR_IMPORT_SHARED_LIBRARY
            #ifdef LIB_ERR_USER_ERROR_HANDLER
                #undef LIB_ERR_USER_ERROR_HANDLER
        
            #endif

        #endif 

    #endif

#endif

//
/////////////////////////////////////////////////////////////////////////////////////////////////////





/////////////////////////////////////////////////////////////////////////////////////////////////////
//
// Define library compile target
//

// Set export string
#define LIB_ERR_LIB_EXPORT 

// Target as header files
#ifdef LIB_ERROR_HEADER_LIBRARY
    #ifdef LIB_ERROR_IMPORT_SHARED_LIBRARY
        #undef LIB_ERROR_IMPORT_SHARED_LIBRARY
    
    #endif
    #ifdef LIB_ERROR_COMPILE_BINARY_LIBRARY
        #undef LIB_ERROR_COMPILE_BINARY_LIBRARY

    #endif

#endif

// Target shared library import
#ifdef LIB_ERROR_IMPORT_SHARED_LIBRARY
    #ifdef LIB_ERROR_HEADER_LIBRARY
        #undef LIB_ERROR_HEADER_LIBRARY
    
    #endif
    #ifdef LIB_ERROR_COMPILE_BINARY_LIBRARY
        #undef LIB_ERROR_COMPILE_BINARY_LIBRARY
    
    #endif

#endif

// Target compile shared library
#ifdef LIB_ERROR_COMPILE_BINARY_LIBRARY
    #ifdef LIB_ERROR_HEADER_LIBRARY
        #undef LIB_ERROR_HEADER_LIBRARY
    
    #endif
    #ifdef LIB_ERROR_IMPORT_SHARED_LIBRARY
        #undef LIB_ERROR_IMPORT_SHARED_LIBRARY
    
    #endif
    #ifdef LIB_ERR_USER_ERROR_HANDLER
        #undef LIB_ERR_USER_ERROR_HANDLER

    #endif
    #undef LIB_ERR_LIB_EXPORT
    #define LIB_ERR_LIB_EXPORT export
    #property library
    
#endif

// Default to header mode
#ifndef LIB_ERROR_HEADER_LIBRARY
    #ifndef LIB_ERROR_IMPORT_SHARED_LIBRARY
        #ifndef LIB_ERROR_COMPILE_BINARY_LIBRARY
            #define LIB_ERROR_HEADER_LIBRARY
            #ifdef LIB_ERR_LIB_EXPORT
                #undef LIB_ERR_LIB_EXPORT
                #define LIB_ERR_LIB_EXPORT
        
            #endif
        
        #endif 
        
    #endif 
    
#endif

//
/////////////////////////////////////////////////////////////////////////////////////////////////////





////////////////////////////////////////////////////
//
// Maximum possible user error code
//
#ifndef ERR_USER_ERROR_LAST
    #define ERR_USER_ERROR_LAST                 0xFFFE

#endif

//
////////////////////////////////////////////////////





////////////////////////////////////////////////////
//
// Define error library identifyer
//
#ifndef LIB_ERROR_LIBRARY_PREFIX
    #define LIB_ERROR_LIBRARY_PREFIX            LibError
    #define LIB_ERROR_STANDALONE_COMPILE

#endif

//
////////////////////////////////////////////////////





////////////////////////////////////////////////////
//
// MQL4 namespace compatibility
//
#ifdef __MQL5__
    #ifndef LIB_NAMESPACE
        #define LIB_NAMESPACE(x, y)             x::y

    #endif
    #ifndef LIB_NAMESPACE_DEF
        #define LIB_NAMESPACE_DEF(x, y)         y

    #endif
#else
    #ifndef __MQL4_COMPATIBILITY_CODE__
        #define __MQL4_COMPATIBILITY_CODE__

    #endif
    #ifndef LIB_NAMESPACE
        #define LIB_NAMESPACE(x, y)             x##_##y

    #endif
    #ifndef LIB_NAMESPACE_DEF
        #define LIB_NAMESPACE_DEF(x, y)         LIB_NAMESPACE(x, y)

    #endif
#endif

//
////////////////////////////////////////////////////





////////////////////////////////////////////////////
//
// Lib Error MT-Version identifyer string
//
#ifndef __MQL4_COMPATIBILITY_CODE__
    	#define LIB_ERR_MT_VERSION_ID_STRING    mt5

#else
    	#define LIB_ERR_MT_VERSION_ID_STRING    mt4

#endif

//
////////////////////////////////////////////////////





////////////////////////////////////////////////////
//
// Terminal language IDs
//

#define LIB_ERROR_LANG_FROM_FILE                0x454C4946          // Read error description from file
#define LIB_ERROR_LANG_TERM_ID_EN               0x6C676E45          // Default language English
#define LIB_ERROR_LANG_TERM_ID_RU               0x73737552          // Russian
#ifndef __MQL4_COMPATIBILITY_CODE__
    #define LIB_ERROR_LANG_TERM_ID_ZH           0x6E696843          // Chinese
    #define LIB_ERROR_LANG_TERM_ID_ES           0x6E617053          // Spanish
    #define LIB_ERROR_LANG_TERM_ID_PT           0x74726F50          // Portugese
    #define LIB_ERROR_LANG_TERM_ID_JP           0x6170614A          // Japanese
    #define LIB_ERROR_LANG_TERM_ID_DE           0x6D726547          // German
    #define LIB_ERROR_LANG_TERM_ID_KO           0x65726F4B          // Korean
    #define LIB_ERROR_LANG_TERM_ID_FR           0x6E657246          // Frensh
    #define LIB_ERROR_LANG_TERM_ID_IT           0x6C617449          // Italian
    #define LIB_ERROR_LANG_TERM_ID_TR           0x6B727554          // Turkish

#endif

//
////////////////////////////////////////////////////





/////////////////////////////////////////////////////////////////////////////////////////////////////
//
//  Custom error code definition macros
//


////////////////////////////////////////////////////
//
// Function name macro substitutions
//

// Select appropiate error handler
#define OnErrorFunction                         DefaultOnError
#ifdef LIB_ERR_USER_ERROR_HANDLER
    #undef OnErrorFunction
    #define OnErrorFunction                     UserOnError

#endif

//
////////////////////////////////////////////////////





////////////////////////////////////////////////////
//
// Define error resolver function pointer
//

// User supplied external error resolver function
typedef const string (*TUsrErrFunc)(const int);

//
////////////////////////////////////////////////////





////////////////////////////////////////////////////
//
// User supplied error codes
//

// Helper macro for defining error codes
#define LIB_ERROR_CODE(x, y)                                                        (uint)(((ushort)((ushort)((uchar)(x) * LIB_ERR_GROUP_SIZE) + (uchar)(y)) + ERR_USER_ERROR_FIRST))
#define LIB_ERROR_CODE_WARNING(x, y)                                                LIB_ERROR_CODE(x, (y + LIB_ERR_LEVEL_WARNING))
#define LIB_ERROR_CODE_HANDLED(x, y)                                                LIB_ERROR_CODE(x, (y + LIB_ERR_LEVEL_HANDLED_ERROR))
#define LIB_ERROR_CODE_GROUP_OFFSET(x)                                              (x << 16)

// Message resolver helper macros
#ifndef __MQL4_COMPATIBILITY_CODE__
    #define LIB_ERR_REGISTER_RESOLVER_BEGIN(_id_)                                   namespace lib_err { const string errdesc_##_id_(const int error_code) { switch(error_code) {
    #define LIB_ERR_REGISTER_RESOLVER_END(_id_)                                     }; return(NULL); } const static bool init##_id_ = LIB_NAMESPACE(LIB_ERROR_LIBRARY_PREFIX, RegisterErrorCodeResolver)(_id_, errdesc_##_id_); }

#else
    #define LIB_ERR_REGISTER_RESOLVER_BEGIN(_id_)                                   const string lib_err_errdesc_##_id_(const int error_code) { switch(error_code) {
    #define LIB_ERR_REGISTER_RESOLVER_END(_id_)                                     }; return(NULL); } const static bool lib_err_init##_id_ = LIB_NAMESPACE(LIB_ERROR_LIBRARY_PREFIX, RegisterErrorCodeResolver)(_id_, lib_err_errdesc_##_id_);

#endif

//
////////////////////////////////////////////////////





////////////////////////////////////////////////////
//
// Reserved error codes
//

// Helper macro for defining reserved error codes
#define LIB_ERROR_RESERVED_CODE(x, y)                                               (uint)(((ushort)(( (ushort)((uchar)(x) * LIB_ERR_GROUP_SIZE) + (uchar)(y))) + LIB_ERR_INTERNAL_GROUP_IDS + ERR_USER_ERROR_FIRST))
#define LIB_ERROR_RESERVED_CODE_WARNING(x, y)                                       LIB_ERROR_RESERVED_CODE(x, (y + LIB_ERR_LEVEL_WARNING))
#define LIB_ERROR_RESERVED_CODE_HANDLED(x, y)                                       LIB_ERROR_RESERVED_CODE(x, (y + LIB_ERR_LEVEL_HANDLED_ERROR))
#define LIB_ERROR_RESERVED_CODE_GROUP_OFFSET(x)                                     (x << 24)

// Message resolver helper macros
#ifndef __MQL4_COMPATIBILITY_CODE__
    #define LIB_ERR_RESERVED_REGISTER_RESOLVER_BEGIN(_id_)                          namespace lib_err { const string errdesc_reserved_##_id_(const int error_code) { switch(error_code) {
    #define LIB_ERR_RESERVED_REGISTER_RESOLVER_END(_id_)                            }; return(NULL); } const static bool init_reserved##_id_ = LIB_NAMESPACE(LIB_ERROR_LIBRARY_PREFIX, RegisterErrorCodeResolver)((LIB_ERR_INTERNAL_GROUP_IDS >> 0x08) + _id_, errdesc_reserved_##_id_); }

#else
    #define LIB_ERR_RESERVED_REGISTER_RESOLVER_BEGIN(_id_)                          const string lib_err_errdesc_reserved_##_id_(const int error_code) { switch(error_code) {
    #define LIB_ERR_RESERVED_REGISTER_RESOLVER_END(_id_)                            }; return(NULL); } const static bool lib_err_init_reserved##_id_ = LIB_NAMESPACE(LIB_ERROR_LIBRARY_PREFIX, RegisterErrorCodeResolver)((LIB_ERR_INTERNAL_GROUP_IDS >> 0x08) + _id_, lib_err_errdesc_reserved_##_id_);

#endif

//
////////////////////////////////////////////////////





////////////////////////////////////////////////////
//
// Error code to text resolver definition
//

#ifndef LIB_ERR_REGISTER_RESOLVER_MSGCODE
    #define LIB_ERR_REGISTER_RESOLVER_MSGCODE(_code_id_, _code_name_, _msg_)        case _code_id_: return(LIB_ERR_OUTPUT_FORMAT(error_code, _code_name_, _msg_));

#endif

//
////////////////////////////////////////////////////





////////////////////////////////////////////////////
//
// Include mql defined error codes and messages
//

// Setup language function macro
#define LIB_ERROR_LANGUAGE_FUNCTION(_lng_code_) lib_err_code_to_text_##_lng_code_

// Include MQL defined error codes
#ifdef LIB_ERROR_HEADER_LIBRARY
    #define LIB_ERROR_INCLUDE_ERROR_TEXTFILE
    
#endif 
#ifdef LIB_ERROR_COMPILE_BINARY_LIBRARY
    #define LIB_ERROR_INCLUDE_ERROR_TEXTFILE
    
#endif 

#ifdef LIB_ERROR_INCLUDE_ERROR_TEXTFILE
    #ifndef __MQL4_COMPATIBILITY_CODE__
        #include <mqlplus/lib_error/lib_errtxt_mql5_error_codes.mqh>
        namespace lib_err { #include <mqlplus/lib_error/lib_errtxt_mql5.mqh> };

    #else
        #include <mqlplus/lib_error/lib_errtxt_mql4_error_codes.mqh>
        #include <mqlplus/lib_error/lib_errtxt_mql4.mqh>

    #endif 
#endif

#ifdef LIB_ERROR_INCLUDE_ERROR_TEXTFILE
    #undef LIB_ERROR_INCLUDE_ERROR_TEXTFILE

#endif

//
////////////////////////////////////////////////////


//
/////////////////////////////////////////////////////////////////////////////////////////////////////






////////////////////////////////////////////////////
//
//  Function import / library inclusion
//
//

#ifdef LIB_ERROR_IMPORT_SHARED_LIBRARY
    #include "lib_error_functions.mqh"

#else
    #include "lib_error_main.mqh"

#endif

//
////////////////////////////////////////////////////





/////////////////////////////////////////////////////////////////////////////////////////////////////
//
// Error handler function overloads
//
// Namespace LIB_ERROR_LIBRARY_PREFIX:: BEGIN
#ifndef __MQL4_COMPATIBILITY_CODE__
    namespace LIB_ERROR_LIBRARY_PREFIX {

#endif


//+------------------------------------------------------------------+
//| OnError()                                                        |
//+------------------------------------------------------------------+
bool LIB_NAMESPACE_DEF(LIB_ERROR_LIBRARY_PREFIX, OnErrorFunction)(const int line_no, const string prefix)
{
    // Local init
    uint err_code = NULL;

    // Return
    return(LIB_NAMESPACE(LIB_ERROR_LIBRARY_PREFIX, OnErrorFunction)(err_code, StringFormat("(Line: %i) %s\r\n", line_no, prefix)));
}


//+------------------------------------------------------------------+
//| OnError()                                                        |
//+------------------------------------------------------------------+
bool LIB_NAMESPACE_DEF(LIB_ERROR_LIBRARY_PREFIX, OnErrorFunction)(const int line_no)
{
    // Local init
    uint err_code = NULL;

    // Return
    return(LIB_NAMESPACE(LIB_ERROR_LIBRARY_PREFIX, OnErrorFunction)(err_code, ::IntegerToString(line_no)));
}


//+------------------------------------------------------------------+
//| OnError()                                                        |
//+------------------------------------------------------------------+
bool LIB_NAMESPACE_DEF(LIB_ERROR_LIBRARY_PREFIX, OnErrorFunction)(const string prefix = NULL)
{
    // Local init
    uint err_code = NULL;

    // Return
    return(LIB_NAMESPACE(LIB_ERROR_LIBRARY_PREFIX, OnErrorFunction)(err_code, prefix));
}

// Namespace handler
#ifndef __MQL4_COMPATIBILITY_CODE__
    }; // Namespace LIB_ERROR_LIBRARY_PREFIX::

#endif


//+------------------------------------------------------------------+
//| OnError()                                                        |
//+------------------------------------------------------------------+
#ifndef LIB_ERR_USER_ERROR_HANDLER
    #ifdef LIB_ERROR_IMPORT_SHARED_LIBRARY
        #ifdef __MQL4_COMPATIBILITY_CODE__
            #ifndef LIB_ERROR_STANDALONE_COMPILE
                #import "LibMQLplus/MQLplus.ex4"
                
            #else 
                #import "LibError/LibError.ex4"
    
            #endif
        
        #else
            #ifndef LIB_ERROR_STANDALONE_COMPILE
                #import "LibMQLplus/MQLplus.ex5"
                
            #else 
                #import "LibError/LibError.ex5"
    
            #endif

        #endif

        const bool LIB_NAMESPACE_DEF(LIB_ERROR_LIBRARY_PREFIX, OnErrorFunction)(uint& err_no, const string prefix = NULL);
        #import

    #endif

#endif

//
/////////////////////////////////////////////////////////////////////////////////////////////////////





/////////////////////////////////////////////////////////////////////////////////////////////////////
//
// Function name macro substitutions
//

// Check for predefined OnError handler
#ifdef OnError
    #undef OnError

#endif

#ifdef LIB_ERROR_IMPORT_SHARED_LIBRARY
    #define err_SetUserError 		                LIB_NAMESPACE(LIB_ERROR_LIBRARY_PREFIX, SetUserErrorCode)
    #define err_ResetLastError                      LIB_NAMESPACE(LIB_ERROR_LIBRARY_PREFIX, ResetUserErrorCode)
    #define err_GetLastError		                LIB_NAMESPACE(LIB_ERROR_LIBRARY_PREFIX, GetLastErrorCode)
    #define err_SetMqlError				            LIB_NAMESPACE(LIB_ERROR_LIBRARY_PREFIX, SetMqlErrorCode)
    #define err_ResetUserError			            LIB_NAMESPACE(LIB_ERROR_LIBRARY_PREFIX, ResetUserErrorCode)
    #define err_ResetUserErrorCode                  LIB_NAMESPACE(LIB_ERROR_LIBRARY_PREFIX, ResetUserErrorCode)
    #define err_ResolveLastError		            LIB_NAMESPACE(LIB_ERROR_LIBRARY_PREFIX, ResolveLastErrorCode)
    #define err_ResolveLastErrorCode		        LIB_NAMESPACE(LIB_ERROR_LIBRARY_PREFIX, ResolveLastErrorCode)
    #define err_GetLastResolvedError                LIB_NAMESPACE(LIB_ERROR_LIBRARY_PREFIX, GetLastResolvedErrorCode)
    #define err_OnError                             LIB_NAMESPACE(LIB_ERROR_LIBRARY_PREFIX, OnErrorFunction)

#endif

#ifdef LIB_ERROR_API_FUNCTION_SHADOWING
    #ifdef SetUserError
	    #undef SetUserError

    #endif
    #ifdef ResetLastError
	    #undef ResetLastError

    #endif
    #ifdef GetLastError
	    #undef GetLastError

    #endif
	#define SetUserError 			        LIB_NAMESPACE(LIB_ERROR_LIBRARY_PREFIX, SetUserErrorCode)
	#define ResetLastError 			        LIB_NAMESPACE(LIB_ERROR_LIBRARY_PREFIX, ResetUserErrorCode)
	#define GetLastError			        LIB_NAMESPACE(LIB_ERROR_LIBRARY_PREFIX, GetLastErrorCode)

#endif
#define SetMqlError				            LIB_NAMESPACE(LIB_ERROR_LIBRARY_PREFIX, SetMqlErrorCode)
#define ResetUserError			            LIB_NAMESPACE(LIB_ERROR_LIBRARY_PREFIX, ResetUserErrorCode)
#define ResolveLastError		            LIB_NAMESPACE(LIB_ERROR_LIBRARY_PREFIX, ResolveLastErrorCode)
#define GetLastResolvedError                LIB_NAMESPACE(LIB_ERROR_LIBRARY_PREFIX, GetLastResolvedErrorCode)
#define OnError                             LIB_NAMESPACE(LIB_ERROR_LIBRARY_PREFIX, OnErrorFunction)

//
/////////////////////////////////////////////////////////////////////////////////////////////////////
/**************************************/
#endif // LIB_ERROR_HANDLER_DEFINITIONS_MQH_INCLUDED
